/*
 * hrtime.c
 *
 * Copyright (C) 2000, Nick Rasmussen <nick@jive.org>
 *
 */
#include "hrtime.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <errno.h>
#include <stdlib.h>

#define HRTIME_HAS_OFFSETS 0
#define PAGE_SIZE	4096

struct hrtime_struct {
    int fd;
    volatile const struct vperfctr_state *kstate;
    unsigned char have_rdpmc;
    unsigned char have_rdtsc;
};

#define rdtsc(low,high) \
     __asm__ __volatile__("rdtsc" : "=a" (low), "=d" (high))

#define rdtscl(low) \
     __asm__ __volatile__("rdtsc" : "=a" (low) : : "edx" )

static inline void get_current_hrtime(volatile hrtime_t *t)
{
        volatile unsigned long *parts = (volatile unsigned long *) t;

        rdtsc(parts[0], parts[1]);
        return;
}

static inline void get_current_hrtime_low(volatile unsigned int *t)
{
        rdtscl(*t);
        return;
}


static struct hrtime_struct *current = 0;

#define HRTIME_PATH_LENGTH 256

int
hrtime_is_present(void)
{
    struct stat tmp;

    if (stat("/proc/self/perfctr", &tmp) == 0)
        return 1;

    return 0;
}

int
hrtime_init(void)
{
    if (current != 0) return 0;
    return get_hrtime_struct(0, &current);
}

#if 0
int
hrtime_uninit(void)
{
    int error;
    if (current == 0) return 0;
    error = free_hrtime_struct(current);
    current = 0;
    return error;
}
#endif

int
get_hrtime_struct(pid_t pid, struct hrtime_struct **dest)
{
    char path[HRTIME_PATH_LENGTH];
    int  count;
    struct hrtime_struct *tmp;
    struct perfctr_info info;
    struct perfctr_control control;
	
    if (dest == 0)
	return -1;

    if (pid == 0) pid = getpid();

    tmp = (struct hrtime_struct *) malloc(sizeof(struct hrtime_struct));
    if (!tmp)
	return -1;

    count = snprintf(path, HRTIME_PATH_LENGTH, "/proc/%d/perfctr", pid);
    if (count < 0 || count >= HRTIME_PATH_LENGTH)
        return -1;
	
    printf("open\n");
    tmp->fd = open(path, O_RDONLY);
    if (tmp->fd < 0)
    {
	int fd = tmp->fd;
	free(tmp);
        return fd;
    }

    printf("ioctl info\n");
    if (ioctl(tmp->fd, PERFCTR_INFO, &info) != 0)
    {
	free(tmp);
	return -1;
    }

    tmp->have_rdpmc = (info.cpu_features & PERFCTR_FEATURE_RDPMC) != 0;
    /* XXX: On VIA Cyrix III rdpmc(0) doesn't work as we want. Not really
       a CPU bug, but it destroys our implicit counter -> tsc/pmc mapping.
       Instead of slowing down the read procedures with special cases,
       pretend we don't have rdpmc at all and always get the accumulated
       values from the kernel (which gets it right). */
    if( info.cpu_type == PERFCTR_X86_VIA_CYRIX_III )
	tmp->have_rdpmc = 0;
    tmp->have_rdtsc = (info.cpu_features & PERFCTR_FEATURE_RDTSC) != 0;

    if (!tmp->have_rdtsc) {
	free(tmp);
	return -1;
    }

    printf("mmap\n");
    tmp->kstate = mmap(NULL, PAGE_SIZE, PROT_READ, MAP_SHARED, tmp->fd, 0);
	
    if (tmp->kstate == MAP_FAILED)
    {
        tmp->kstate = 0;
	free(tmp);
        return errno;
    }

    // 'control' currently has the effect of starting the counter and
    // resetting the tsc start time
    printf("control\n");
    memset(&control, 0, sizeof control);
    if( ioctl(tmp->fd, VPERFCTR_CONTROL, &control) < 0 ) {
        free(tmp);
	return -1;
    }


    *dest = tmp;

    return 0;
} 

int
free_hrtime_struct(struct hrtime_struct *hr)
{
    int error;

    if (hr == 0) return 0;
    
    error = munmap((void *) hr->kstate, PAGE_SIZE);
    if (error != 0)
        return error;

    return 0;
}

#define HRTIME_CYCLE_LIMIT 100

void
get_hrvtime_self(hrtime_t *dest)
{
    unsigned int ld, cur;
    hrtime_t vtime;

    do {
	ld        = current->kstate->start.ctr[0];
	vtime     = current->kstate->sum.ctr[0];
	get_current_hrtime_low(&cur);
    }
    while (ld != current->kstate->start.ctr[0]);

    *dest = vtime + (cur - ld);
}

void
get_hrvtime(struct hrtime_struct *hr, hrtime_t *dest)
{
    unsigned int cur, ld, my_ld;
    hrtime_t vtime;
    int status;

#if HRTIME_HAS_OFFSETS
    long     my_off, hr_off;
#endif

    do {
	ld        = hr->kstate->start.ctr[0];
	my_ld     = current->kstate->start.ctr[0];
	vtime     = hr->kstate->sum.ctr[0];
	status	  = hr->kstate->status;
#if HRTIME_HAS_OFFSETS
	my_off    = current->offset_to_cpu0;
	hr_off    = hr->offset_to_cpu0;
#endif
	get_current_hrtime_low(&cur);
    }
    while (ld != hr->kstate->start.ctr[0] || my_ld != current->kstate->start.ctr[0]);

    if (status != VPERFCTR_STATUS_STOPPED)
    {
#if HRTIME_HAS_OFFSETS
	/* make ld in reference to our cpu */
	ld += (hr_off - my_off);
#endif
	*dest = vtime + (cur - ld);
    }
    else
    {
        /* queried process is not on a cpu, vtime is correct */
	*dest = vtime;
    }
}

